/*!
 * \file   src/System/System.cxx
 * \brief    
 * \author Helfer Thomas
 * \date   06 Nov 2007
 * \copyright Copyright (C) 2006-2014 CEA/DEN, EDF R&D. All rights 
 * reserved. 
 * This project is publicly released under either the GNU GPL Licence 
 * or the CECILL-A licence. A copy of thoses licences are delivered 
 * with the sources of TFEL. CEA or EDF may also distribute this 
 * project under specific licensing conditions. 
 */
#include<iostream>

#include<sstream>
#include<fstream>
#include<algorithm>
#include<functional>
#include<cstdio>
#include<cstring>
#include<cstdlib>
#include<climits>

#include<unistd.h>
#include<dirent.h>

#if defined _WIN32 || defined _WIN64 ||defined __CYGWIN__
#include<windows.h>
#endif 

#include"TFEL/System/SystemError.hxx"
#include"TFEL/System/System.hxx"

namespace tfel
{

  namespace system
  {

    char
    dirSeparator(void)
    {
      return '/';
    } // end of systemCall::dirSeparator

    const std::string&
    dirStringSeparator(void)
    {
      using namespace std;
      static string s(1u,dirSeparator());
      return s;
    } // end of systemCall::dirStringSeparator

    void
    systemCall::throwSystemError(const std::string& errMsg,
				 const int errNbr)
    {
      using namespace std;
      string err = strerror(errNbr);
      transform(err.begin(),err.end(),err.begin(),ptr_fun(::tolower));
      string msg(errMsg);
      msg += " (system error : ";
      msg += err;
      msg += ")";
      switch(errNbr){
      case EPERM:
	throw(PosixError<EPERM>(msg));
	break;
      case ENOENT:
	throw(PosixError<ENOENT>(msg));
	break;
      case ESRCH:
	throw(PosixError<ESRCH>(msg));
	break;
      case EINTR:
	throw(PosixError<EINTR>(msg));
	break;
      case EIO:
	throw(PosixError<EIO>(msg));
	break;
      case ENXIO:
	throw(PosixError<ENXIO>(msg));
	break;
      case E2BIG:
	throw(PosixError<E2BIG>(msg));
	break;
      case ENOEXEC:
	throw(PosixError<ENOEXEC>(msg));
	break;
      case EBADF:
	throw(PosixError<EBADF>(msg));
	break;
      case ECHILD:
	throw(PosixError<ECHILD>(msg));
	break;
      case EAGAIN:
	throw(PosixError<EAGAIN>(msg));
	break;
      case ENOMEM:
	throw(PosixError<ENOMEM>(msg));
	break;
      case EACCES:
	throw(PosixError<EACCES>(msg));
	break;
      case EFAULT:
	throw(PosixError<EFAULT>(msg));
	break;
      case EBUSY:
	throw(PosixError<EBUSY>(msg));
	break;
      case EEXIST:
	throw(PosixError<EEXIST>(msg));
	break;
      case EXDEV:
	throw(PosixError<EXDEV>(msg));
	break;
      case ENODEV:
	throw(PosixError<ENODEV>(msg));
	break;
      case ENOTDIR:
	throw(PosixError<ENOTDIR>(msg));
	break;
      case EISDIR:
	throw(PosixError<EISDIR>(msg));
	break;
      case EINVAL:
	throw(PosixError<EINVAL>(msg));
	break;
      case ENFILE:
	throw(PosixError<ENFILE>(msg));
	break;
      case EMFILE:
	throw(PosixError<EMFILE>(msg));
	break;
      case ENOTTY:
	throw(PosixError<ENOTTY>(msg));
	break;
      case EFBIG:
	throw(PosixError<EFBIG>(msg));
	break;
      case ENOSPC:
	throw(PosixError<ENOSPC>(msg));
	break;
      case ESPIPE:
	throw(PosixError<ESPIPE>(msg));
	break;
      case EROFS:
	throw(PosixError<EROFS>(msg));
	break;
      case EMLINK:
	throw(PosixError<EMLINK>(msg));
	break;
      case EPIPE:
	throw(PosixError<EPIPE>(msg));
	break;
      case EDOM:
	throw(PosixError<EDOM>(msg));
	break;
      case ERANGE:
	throw(PosixError<ERANGE>(msg));
	break;
#if not (defined _WIN32 || defined _WIN64 ||defined __CYGWIN__)
      case ENOTBLK:
	throw(PosixError<ENOTBLK>(msg));
	break;
      case ETXTBSY:
	throw(PosixError<ETXTBSY>(msg));
	break;
#endif /* defined _WIN32 || _WIN64 || defined __CYGWIN__ */      
      default:
	throw(SystemError(msg));
      }
    } // end of throwSystemError

    std::vector<std::string>
    systemCall::tokenize(const std::string& s,
			 const char c)
    {
      using namespace std;
      vector<string> res;
      string::size_type b = 0u;
      string::size_type e = s.find_first_of(c, b);
      while (string::npos != e || string::npos != b){
	// Found a token, add it to the vector.
	res.push_back(s.substr(b, e - b));
	b = s.find_first_not_of(c, e);
	e = s.find_first_of(c, b);
      }
      return res;
    } // end of systemCall::tokenize

    std::vector<std::string>
    systemCall::tokenize(const std::string& s,
			 const std::string& delim)
    {
      using namespace std;
      vector<string> res;
      string::size_type b = 0u;
      string::size_type e = s.find_first_of(delim, b);
      while (string::npos != e || string::npos != b){
	// Found a token, add it to the vector.
	res.push_back(s.substr(b, e - b));
	b = s.find_first_not_of(delim, e);
	e = s.find_first_of(delim, b);
      }
      return res;
    } // end of systemCall::tokenize

    void
    systemCall::copy(const std::string& src,
		     const std::string& dest)
    {
      using namespace std;
      struct stat srcInfos;
      struct stat destInfos;
      bool destStatus;
      if(src==dest){
	return;
      }
      if(::stat(src.c_str(),&srcInfos)==-1){
	systemCall::throwSystemError("systemCall::copy : can't stat file "+src,errno);
      }
      destStatus=::stat(dest.c_str(),&destInfos)==0;
      if(!destStatus){
	if(errno!=ENOENT){
	  systemCall::throwSystemError("systemCall::copy : can't stat file "+dest,errno);
	}
      }
      if(S_ISDIR(srcInfos.st_mode)){
	// copying a directory
	if(destStatus){
	  // the destination exist, it shall be a directory
	  if(!S_ISDIR(destInfos.st_mode)){
	    throw(SystemError("systemCall::copy : can't copy directory '"+
			      src+"' on "+systemCall::fileType(destInfos.st_mode)
			      +" '"+dest+"'."));
	  }
	  systemCall::copyDirectory(src,dest,true);
	  return;
	} else {
	  systemCall::copyDirectory(src,dest,false);
	  return;
	}
      } else if(S_ISREG(srcInfos.st_mode)){
	// copying a file, the destination
	// shall be a file or a directory
	if(destStatus){
	  if(S_ISDIR(destInfos.st_mode)){
	    // destination is a directory
	    const vector<string>& tmp = systemCall::tokenize(src,'/');
	    systemCall::copyFile(src,dest+'/'+tmp.back());
	    return;
	  } else if (S_ISREG(destInfos.st_mode)){
	    // destination is a file
	    systemCall::copyFile(src,dest);
	    return;
	  }
	  throw(SystemError("systemCall::copy : can't copy file '"+src+"' on "+
			    systemCall::fileType(destInfos.st_mode)+" '"+dest+"'."));
	}
	// destination don't exist, trying to copy to it
	systemCall::copyFile(src,dest);
	return;
      } else if(S_ISCHR(destInfos.st_mode)||S_ISBLK(destInfos.st_mode)||
#if defined _WIN32 || defined _WIN64 ||defined __CYGWIN__
#warning "windows port"
#else
		S_ISSOCK(destInfos.st_mode)||S_ISLNK(destInfos.st_mode)||
#endif /* defined _WIN32 || _WIN64 || defined __CYGWIN__ */      
		S_ISFIFO(destInfos.st_mode)){
	string msg("systemCall::copy : can't copy object of type "+
		   systemCall::fileType(destInfos.st_mode));	
	throw(SystemError(msg));
      }
    } // end of systemCall::copy

#if defined _WIN32 || defined _WIN64 ||defined __CYGWIN__
#warning "windows port"
    void
    systemCall::mkdir(const std::string& dir)
#else 
    void
    systemCall::mkdir(const std::string& dir,const mode_t mode)
#endif /* defined _WIN32 || _WIN64 || defined __CYGWIN__ */      
    {
      using namespace std;
      string path;
      const vector<string>& paths = systemCall::tokenize(dir,'/');
      vector<string>::const_iterator p;
      for(p=paths.begin();p!=paths.end();++p){
	if(!path.empty()){
	  path+="/"+*p;	  
	} else {
	  path=*p;
	}
	struct stat infos;
	if(::stat(path.c_str(),&infos)==-1){
	  if(errno==ENOENT){
	    // the file does not exist, create the directory
#if defined _WIN32 || defined _WIN64 ||defined __CYGWIN__
#warning "windows port"
	    if(::CreateDirectoryA(path.c_str(),0)==0){
#else 
	    if(::mkdir(path.c_str(),mode)!=0){
#endif /* defined _WIN32 || _WIN64 || defined __CYGWIN__ */      
	      if(errno==EEXIST){
		// this may happen if many processes are
		// used in the same time
		if(::stat(path.c_str(),&infos)==0){
		  if(S_ISDIR(infos.st_mode)){
		    return;
		  }
		} else {
		  systemCall::throwSystemError("systemCall::mkdir : can't stat file '"+*p+"'",errno);
		}
	      }
	      systemCall::throwSystemError("systemCall::mkdir : can't create directory '"+dir+"'",errno);
	    }
	  } else {
	    systemCall::throwSystemError("systemCall::mkdir : can't stat file '"+*p+"'",errno);
	  }
	} else {
	  // the path is valid, check if it is a directory
	  if(!S_ISDIR(infos.st_mode)){
	    string msg("systemCall::mkdir : '");
	    msg += path+"' exists and is not a directory.";
	    throw(SystemError(msg));
	  }
	}
      }
    } // end of systemCall::mkdir
  
    void
    systemCall::unlink(const std::string& f)
    {
      if(::unlink(f.c_str())!=0){
	systemCall::throwSystemError("systemCall::unlink : can't unlink file "+f,errno);
      }
    } // end of systemCall::unlinkx

    void
    systemCall::rmdir(const std::string& d)
    {
      using namespace std;
      struct stat dInfos;
      vector<string> o;
      if(::stat(d.c_str(),&dInfos)==-1){
	systemCall::throwSystemError("systemCall::rmdir : can't stat directory "+d,errno);
      }
      if(!S_ISDIR(dInfos.st_mode)){
	string msg("systemCall::rmdir : '");
	msg += d+"' is not a directory";
	throw(runtime_error(msg));
      }
      DIR *dir = opendir(d.c_str());
      struct dirent *p;
      if(dir==0){
	systemCall::throwSystemError("systemCall::rmdir, can't open directory "+d,
				     errno);
      }
      while((p=readdir(dir))!=NULL){
	if((strcmp(p->d_name,".")!=0)&&
	   (strcmp(p->d_name,"..")!=0)){
	  o.push_back(p->d_name);
	}
      }
      vector<string>::const_iterator po;
      for(po=o.begin();po!=o.end();++po){
	struct stat pInfos;
	string f;
	f = d+"/"+*po;
	if(::stat(f.c_str(),&pInfos)==-1){
	  systemCall::throwSystemError("systemCall::rmdir : can't stat '"+f+"'",errno);
	}
	if(S_ISDIR(pInfos.st_mode)){
	  systemCall::rmdir(f);
	} else {
	  systemCall::unlink(f);
	}
      }
      closedir(dir);
      ::rmdir(d.c_str());
    } // end of systemCall::rmdir

    void
    systemCall::write(const int f,
		      const void* const v,
		      size_t s)
    {
      /*
       * code adapated from C. Blaess
       * "Programmation systme en C sous linux"
       * 2me dition, 2005, Eyrolles
       * page 5009
       */ 
      using namespace std;
      const unsigned char * b = static_cast<const unsigned char*>(v);
      size_t  r = s;
      ssize_t w;
      if(s>=SSIZE_MAX){
	string msg("systemCall::write : invalid size (s>=SSIZE_MAX)");
	throw(SystemError(msg));
      }
      while(r>0){
	while((w=::write(f,b,r))==-1){
	  if(errno==EINTR){
	    continue;
	  }
	  if(errno!=EAGAIN){
	    systemCall::throwSystemError("systemCall::write failed",errno);
	  }
#if defined _WIN32 || defined _WIN64 ||defined __CYGWIN__
#warning "windows port"
#else
	  ::sleep(1);
#endif /* defined _WIN32 || _WIN64 || defined __CYGWIN__ */      
	}
	r  = static_cast<size_t>(r-w);
	b += w;
      }
    } // end of systemCall::write

    std::string
    systemCall::fileType(const mode_t mode)
    {
      using namespace std;
      if(S_ISDIR(mode)){
	return "directory";
      } else if(S_ISREG(mode)){
	return "regular file";
      } else if(S_ISCHR(mode)){
	return "character device";
      } else if(S_ISBLK(mode)){
	return "block device";
      } else if(S_ISFIFO(mode)){
	return "fifo";
#if defined _WIN32 || defined _WIN64 ||defined __CYGWIN__
#warning "windows port"
#else
      } else if(S_ISLNK(mode)){
	return "symbolic link";
      } else if(S_ISSOCK(mode)){
	return "socket";
#endif /* defined _WIN32 || _WIN64 || defined __CYGWIN__ */      
      }
      return "unknown";
    } // end of systemCall::fileType

    void
    systemCall::copyDirectory(const std::string& src,
			      const std::string& dest,
			      const bool exists)
    {
      using namespace std;
      const vector<string>& paths = systemCall::tokenize(src,'/');
      string rdest;
      if(exists){
	rdest=dest+'/'+paths.back();
      } else {
	rdest=dest;
      }
      systemCall::mkdir(rdest);
      DIR *dir = opendir(src.c_str());
      struct dirent *p;
      if(dir==0){
	systemCall::throwSystemError("systemCall::copy, can't open directory "+src,
				     errno);
      }
      while((p=readdir(dir))!=NULL){
	if((strcmp(p->d_name,".")!=0)&&
	   (strcmp(p->d_name,"..")!=0)){
	  systemCall::copy(src+'/'+p->d_name,rdest);
	}
      }
      closedir(dir);
    } // end of systemCall::copyDirectory

    void
    systemCall::copyFile(const std::string& src,const std::string& dest)
    {
      using namespace std;
      struct stat srcInfos;
      if(::stat(src.c_str(),&srcInfos)==-1){
	systemCall::throwSystemError("systemCall::copyFile : can't stat file "+src,errno);
      }
      if(!S_ISREG(srcInfos.st_mode)){
	string msg("systemCall::copyFile : source '"+src+"' is not a regular file");
	throw(runtime_error(msg));
      }
      // simple check
      if(src==dest){
	return;
      }
      // do the copy
      ofstream out(dest.c_str(),ios_base::binary);
      ifstream in(src.c_str(),ios_base::binary);
      if(!in){
	string msg("systemCall::copyFile : can't open file "+src+" in read mode.");
	throw(SystemError(msg));
      }
      if(!out){
	string msg("systemCall::copyFile : can't open file "+dest+" in write mode.");
	throw(SystemError(msg));
      }
      // local copy of the file
      if(srcInfos.st_size!=0){
	// file is not empty
	out.exceptions(ifstream::failbit | ifstream::badbit);
	out << in.rdbuf();
      }
      in.close();
      out.close();
    } // end of systemCall::copyFile
    
    std::string
    systemCall::getCurrentWorkingDirectory(void)
    {
      using namespace std;
      char *name  = 0;
      size_t size = 16u;
      while(1){
	if((name=static_cast<char *>(realloc(name,size)))==0){
	  throw(SystemError("systemCall::getCurrentWorkingDirectory : out of memory"));
	}
	if(::getcwd(name,size)!=0){
	  break;
	}
	if(errno!=ERANGE){
	  systemCall::throwSystemError("systemCall::getCurrentWorkingDirectory ",errno);
	}
	size*=2u;
      }
      const string res(name);
      ::free(name);
      return res;
    } // end of systemCall::getCurrentWorkingDirectory

    std::string
    systemCall::getHostName(void)
    {
      using namespace std;
#if defined _WIN32 || defined _WIN64 ||defined __CYGWIN__
#warning "windows port"
      return "";
#else
      char *name  = 0;
      size_t size = 16u;
      while(1){
	if((name=static_cast<char *>(realloc(name,size)))==0){
	  throw(SystemError("systemCall::getHostName : out of memory"));
	}
	if(::gethostname(name,size)==0){
	  break;
	}
	if(errno!=ENAMETOOLONG){
	  systemCall::throwSystemError("systemCall::getHostName ",errno);
	}
	size*=2u;
      }
      const string res(name);
      ::free(name);
      return res;
#endif
    } // end of systemCall::getHostName

    std::string
    systemCall::getUserName(void)
    {
#if defined _WIN32 || defined _WIN64 ||defined __CYGWIN__
#warning "windows port"
      return "";
#else
      const char * const l = ::getlogin();
      if(l!=0){
	return l;
      }
      return "";
#endif
    } // end of systemCall::getUserName

    void
    systemCall::changeCurrentWorkingDirectory(const std::string& name)
    {
      using namespace std;
      if(::chdir(name.c_str())==-1){
	string msg("systemCall::changeCurrentWorkingDirectory : ");
	msg += "can't change to directory "+name+".";
	systemCall::throwSystemError(msg,errno);
      }
    } // end of systemCall::changeCurrentWorkingDirectory

  } // end of namespace system

} // end of namespace tfel
